/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * definition of a channel (had to put it in a new file because xm.h and mixer.h use it and they would include
 * each other if it would have stayed in mixer.h
 */

#ifndef _CHANNELS_DEF_H_
#define _CHANNELS_DEF_H_

#include "xm.h"

/* the channel structure */
typedef struct {
/*  0 */  instrument *instrument;
/*  4 */  sample *sample;
/*  8 */  unsigned long long sample_position;
/* 16 */  unsigned long long increment;
/* 24 */  int increment_positif;
/* 28 */  int in_use;
/* 32 */  int left_volume;
/* 36 */  int right_volume; /* 0..64 */
	  /* for note delay (EDx) */
/* 40 */  int do_we_delay;

  int volume_point;
  int panning_point;
  int envelop_volume;
  int envelop_panning;
  int volume;
  int panning;
  int cut_key;
  int fade_volume;
  int instrument_vibrato_pos;
  int instrument_vibrato_sweep;
  int closed;	/* put to 1 if the mixer has not to take care of it
		 * (when you put it to 1, don't forget to put in_use to 0)
		 */
  int last_instrument; /* the last instrument put in a channel (init to 0, don't know what has to be) */
  /* those bytes reflect what was in the pattern */
  int note;
  int volume_effect;
  int effect_type;
  int effect_parameter;
  /* some effects will use the last parameter that was given, so we must save them */
  /* take care ! They are never reseted, so you could have a module playing a certain
   * way at the creation time, and when replaying it, it would have another behavior.
   */
  int old_volume_slide;
  int old_fine_volume_slide_up;
  int old_fine_volume_slide_down;
  int old_global_volume_slide;
  int old_panning_slide;
  /* data for e6 effect */
  int loop_point_is_set;
  int looped_nb;
  int loop_row;
  int loop_pattern;
  int loop_max;
  int has_looped;
  /* ok, to deal with vibrato, ..., we need to calculate increment at each tick, so there is the period */
  unsigned long period;
  /* data for portamento up/down */
  int old_porta_up;
  int old_porta_down;
  /* are the two next ok ? shouldn't we use the two above ? */
  int old_fine_porta_up;
  int old_fine_porta_down;
  /* same questions for those two next vars */
  int old_extra_fine_porta_up;
  int old_extra_fine_porta_down;
  /* data for portamento to note */
  int glissendo;	/* do we glissendo ? */
  int old_porta_note;
  int porta_period;	/* the wanted period */
  /* for portamento to note with e31 set */
  int porta_note_one;
  int porta_note_two;
  int porta_half_one;
  int porta_half_two;
  int calc_period;
  int last_no_porta_note;
  /* data for vibrato */
  int vibrato_rate, vibrato_depth;
  int vibrato_control;		/* bit 2 = retrig/not retrig when changing note, 1-0 : vibrato type */
  /* if sample is not over and next row has no vibrato, sample takes the period of the last note, so it's needed */
  int saved_period;
  /* data for tremolo */
  int tremolo_rate, tremolo_depth;
  int tremolo_control;		/* bit 2 = retrig/not retrig when changing note, 1-0 : vibrato type */
  int saved_volume;	/* same as vibrato, except that the volume won't be restored after a tremolo */
  /* for the 0 effect, remember what was the real last notez (not 96 nor -1) */
  int last_note;
  /* to do retrig (e9) */
  int retrig;
  /* the doc says no memorisation, but Ripper says yes, so yes. */
  int last_retrig;
  /* multiple retrig (Rxx) */
  int m_retrig;
  int m_last_retrig;
  int m_volume_effect;
  /* for tremor (Txx) */
  int tremor;
  int old_tremor;
  /* well, those char go here so all the rest can be 4 aligned */
  char fine_tune;
  signed char vibrato_point;		/* because of some mikmod ripped code */
  signed char tremolo_point;
  char __; /* so the length of this structure is 284 bytes long, ie multiple of 4 */
} channel __attribute__ ((packed));

#endif /* _CHANNELS_DEF_H_ */
