/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * the "live"-player functions
 */

#include "live_player.h"

#include <string.h>
#include <stdio.h>

#include "xm.h"
#include "channels_def.h"
#include "card.h"

/* the mix-buffer is in mixer.c */
extern int mixer_soft_buffer[1024*2/4];
extern int mixer_pos;

int get_free_live_player_channel(XM *xm)
{
  int i;
  /* Ok, well, i had a little bug, so i remove this for.
   * When i played a chord, i lost some notes, simply because
   * the same channel was choosed at this place, the in_use can't
   * be put like that, the mixer has to set it after some
   * init. Anyway, it works fine without it. Damned bug.
   */
#if 0
  for (i=0; i<10; i++)
    if (!xm->live_channels[i].in_use)
      return i;
#endif
  i=xm->last_live_channel;
  xm->last_live_channel++;
  if (xm->last_live_channel==10)
    xm->last_live_channel=0;
  return i;
}

int find_live_player_channel(XM *xm, int note)
{
  int i;
  for (i=0; i<10; i++)
    if (xm->live_channels[i].in_use && xm->live_notes[i][1].note==note)
      return i;
  return -1;
}

void put_live_player_note(XM *xm, int channel_number, int note)
{
  /* put the note */
  xm->live_notes[channel_number][0].note=note;
  /* the instrument */
  xm->live_notes[channel_number][0].instrument=xm->start_5+xm->cur_pos_5;
  /* rest is unused (volume=default volume of instrument, no effect) */
  /* we copy the first in the second to remember for later, when keyrealese occurs */
  xm->live_notes[channel_number][1]=xm->live_notes[channel_number][0];
}

void open_live_player(XM *xm)
{
  xm->mix|=LIVE_PLAY;
  if (!init_sound(44100, 16, 1)) {
    xm->mix=0;
    close_live_player(xm);
    fprintf(stderr, "live_player.open_live_player : error with init_sound.\n");
  }
}

void close_live_player(XM *xm)
{
  int i;
  if (!xm->mix)
    return;
  xm->mix&=~LIVE_PLAY;
  if (!xm->mix)
    close_sound();
  for (i=0; i<10; i++)
    xm->live_channels[i].in_use=0;
  memset(xm->live_notes, -1, 20*sizeof(note));
}

/* all the following are the same as in mixer.c, but with only what is needed for "live"-playing (no effect handling) */

/* the following fun is in mixer.c */
unsigned long get_period(instrument *, int, int);
long long get_increment(instrument *, int, unsigned long, int);
extern unsigned char avibtab[128];

int live_new_line(XM *module)
{
  channel *channels=module->live_channels;
  int i;
  note *cur_note;

  /* we new_line at every tick */
  cur_note=&module->live_notes[0][0];

  /* ok, let's run through the notes */
  for (i=0; i<10; i++, cur_note++) {
    /*fprintf(stderr, "n %d i %d\t", cur_note->note, cur_note->instrument);*/
    if (cur_note->note==0xff) {
      cur_note++;
      continue;
    }
    cur_note->note=0xff;
    cur_note++;
    /* test cut-note, if envelop is there with sustain or without it */
    if (cur_note->note==96) {
      if (channels[i].instrument && (channels[i].instrument->volume_type & 1))
	channels[i].cut_key=1;
      else
	channels[i].in_use=0;
      continue;
    }
    /*if (cur_note->instrument==-1)
     *  continue;*/
    /* we load a new note, set all parameters */
    if (!(channels[i].instrument=module->instruments[(int)cur_note->instrument])) {
      channels[i].in_use=0;
      continue;
    }
    channels[i].sample=&channels[i].
      instrument->sample_headers[(int)channels[i].instrument->sample_number[(int)cur_note->note]];
    if (channels[i].sample->sample_data==(char *)0) {
      channels[i].in_use=0;
      continue;
    }
    channels[i].sample_position=0;
    channels[i].increment_positif=1;
    channels[i].period=get_period(module->instruments[(int)cur_note->instrument], cur_note->note, module->flags);
#if 0
    channels[i].increment=get_increment(module->instruments[(int)cur_note->instrument], cur_note->note,
					get_period(module->instruments[(int)cur_note->instrument], cur_note->note));
#endif
    channels[i].in_use=1;
    channels[i].volume_point=0;
    channels[i].panning_point=0;
    channels[i].envelop_volume=channels[i].instrument->volume_type & 1 ? channels[i].instrument->volume_envelop[0][1] : 64;
    channels[i].envelop_panning=channels[i].instrument->panning_type & 1 ?
      channels[i].instrument->panning_envelop[0][1] : 32;
    channels[i].volume=channels[i].sample->sample_volume;
    channels[i].panning=channels[i].sample->sample_panning;
    channels[i].cut_key=0;
    channels[i].fade_volume=32768;
    channels[i].instrument_vibrato_pos=0;
    channels[i].instrument_vibrato_sweep=0;
  }

  /* we've changed the line, let's handle the rest */
  /* let's handle the volume / panning */
  for (i=0; i<10; i++) {
    instrument *ins;

    if (!channels[i].in_use)
      continue;

    ins=channels[i].instrument;

    /* deal with auto vibrato (adapted from mikmod) */
    {
      int vibval=0, vibdepth, insdep=ins->vibrato_depth*4;

      if (insdep)
	switch(ins->vibrato_type) {
	  case 0 :
	    vibval = avibtab[channels[i].instrument_vibrato_pos & 127];
	    if(channels[i].instrument_vibrato_pos & 0x80)
	      vibval=-vibval;
	    break;
	  case 1 :
	    vibval=64;
	    if(channels[i].instrument_vibrato_pos & 0x80)
	      vibval=-vibval;
	    break;
	  case 2 :
	    vibval = 63-(((channels[i].instrument_vibrato_pos + 128) & 255) >> 1);
	    break;
	  case 3 :
	    vibval = (((channels[i].instrument_vibrato_pos + 128) & 255) >> 1) - 64;
	    break;
	}

      if (!channels[i].cut_key) {
	if (channels[i].instrument_vibrato_sweep < ins->vibrato_sweep) {
	  vibdepth=(channels[i].instrument_vibrato_sweep*insdep)/ins->vibrato_sweep;
	  channels[i].instrument_vibrato_sweep++;
	} else
	  vibdepth=insdep;
      } else {
	/* if cut-key, then if to the end of the sweep then depth keeps as it is otherwise becomes 0 */
	if (channels[i].instrument_vibrato_sweep>=ins->vibrato_sweep)
	  vibdepth=insdep;
	else
	  vibdepth=0;
      }
      vibval=(vibval*vibdepth)>>8;
      channels[i].instrument_vibrato_pos+=ins->vibrato_rate;
      channels[i].instrument_vibrato_pos&=255;

      /* ok, calculate the increment too */
      channels[i].increment=get_increment(ins, channels[i].note, channels[i].period-vibval, module->flags);
    }

    /* if cut key, we decrement fade volume */
    if (channels[i].cut_key) {
      if (channels[i].instrument) /* useless test, i think, but i feel better with it */
	channels[i].fade_volume-=ins->volume_fadeout;
      else
	channels[i].fade_volume=0;
      if (channels[i].fade_volume<=0) {
	channels[i].in_use=0;
	continue;
      }
    }

    /* first, set the left/right volume according to the current volume / panning setting */
    {
      /* calculate final volume according to xm.txt */
      int pan=(unsigned char)channels[i].panning;
      int envpan=channels[i].envelop_panning;
      int final_panning=pan + ((envpan-32)*(128-abs(pan-128)))/32;

      channels[i].left_volume=(channels[i].volume*module->global_volume*
			       channels[i].envelop_volume*(255-final_panning))>>(6+6+8);
      channels[i].right_volume=(channels[i].volume*module->global_volume*
				channels[i].envelop_volume*final_panning)>>(6+6+8);
      /* deal with auto fadeout volume */
      channels[i].left_volume=(channels[i].left_volume*channels[i].fade_volume)/32768;
      channels[i].right_volume=(channels[i].right_volume*channels[i].fade_volume)/32768;
    }

    /* now, move the volume point */
    if (!(ins->volume_type & 1))
      goto deal_with_panning;

    /* are we on the sustain point ? */
    if ((!channels[i].cut_key) && (ins->volume_type & 2) &&
	(channels[i].volume_point == ins->volume_envelop[(int)ins->volume_sustain_point][0]))
      goto deal_with_panning;

    /* was it the loop point ? */
    if (ins->volume_type & 4) {
      if (channels[i].volume_point == ins->volume_envelop[(int)ins->volume_loop_end][0]) {
	channels[i].volume_point=ins->volume_envelop[(int)ins->volume_loop_start][0];
	channels[i].envelop_volume=ins->volume_envelop[(int)ins->volume_loop_start][1];
	goto deal_with_panning;
      }
    }

    /* or the end ? */
    if (channels[i].volume_point >= ins->volume_envelop[ins->nb_volume_points-1][0]) {
      /* if so, we start the fade out, as in mikmod */
      channels[i].cut_key=1;
      goto deal_with_panning;
    }

    channels[i].volume_point++;
    /* set the envelop volume (ugly code) */
    {
      int right;
      int x, x1, y1, x2, y2;
      for (right=1; right<ins->nb_volume_points; right++)
	if (ins->volume_envelop[right][0]>=channels[i].volume_point)
	  break;
      /* right is the first point on the right of the current volume point, we interpolate to find the new volume */
      /*
	          /(x2, y2)
		 /
		/ (x, Y)
	       /(x1, y1)
           (x-x1)/(x2-x1)=(Y-y1)/(y2-y1)
	   Y=y1 + (x-x1)/(x2-x1)*(y2-y1)
       */
      if (right < ins->nb_volume_points) {
	x=channels[i].volume_point;
	x1=ins->volume_envelop[right-1][0];
	y1=ins->volume_envelop[right-1][1];
	x2=ins->volume_envelop[right][0];
	y2=ins->volume_envelop[right][1];
	x-=x1;
	x2-=x1;
	y2-=y1;
	channels[i].envelop_volume = y1 + x*y2/x2;
      }
    }

deal_with_panning:
    /* now, move the panning point */
    if (!(ins->panning_type & 1))
      continue;

    /* are we on the sustain point ? */
    if ((!channels[i].cut_key) && (ins->panning_type & 2) &&
	(channels[i].panning_point == ins->panning_envelop[(int)ins->panning_sustain_point][0]))
      continue;

    /* was it the loop point ? */
    if (ins->panning_type & 4) {
      if (channels[i].panning_point == ins->panning_envelop[(int)ins->panning_loop_end][0]) {
	channels[i].panning_point=ins->panning_envelop[(int)ins->panning_loop_start][0];
	channels[i].envelop_panning=ins->panning_envelop[(int)ins->panning_loop_start][1];
	continue;
      }
    }

    /* or the end ? */
    if (channels[i].panning_point >= ins->panning_envelop[ins->nb_panning_points-1][0])
      continue;

    channels[i].panning_point++;
    /* set the envelop panning (ugly code) */
    {
      int right;
      int x, x1, y1, x2, y2;
      for (right=1; right<ins->nb_panning_points; right++)
	if (ins->panning_envelop[right][0]>=channels[i].panning_point)
	  break;
      /* right is the first point on the right of the current panning point, we interpolate to find the new panning */
      /*
	          /(x2, y2)
		 /
		/ (x, Y)
	       /(x1, y1)
           (x-x1)/(x2-x1)=(Y-y1)/(y2-y1)
	   Y=y1 + (x-x1)/(x2-x1)*(y2-y1)
       */
      if (right < ins->nb_panning_points) {
	x=channels[i].panning_point;
	x1=ins->panning_envelop[right-1][0];
	y1=ins->panning_envelop[right-1][1];
	x2=ins->panning_envelop[right][0];
	y2=ins->panning_envelop[right][1];
	x-=x1;
	x2-=x1;
	y2-=y1;
	channels[i].envelop_panning = y1 + x*y2/x2;
      }
      /*      channels[i].panning <<=2;*/
    }
  }

  return (44100L*8*5/(module->bpm*2))&(~7); /*module->nb_bytes;*/
}

void live_handle_one_datum(XM *module)
{
  channel *channels=module->live_channels;
  int i;
  /* for all channels of the module, see if it's currently in use, then take the datum from the sample,
   * and add it to the mixer-buffer with correct left and right volume (and take care if 8 or 16b sample).
   * Then, increment the current pos in the sample, deal with loop or end of sample. */
  for (i=0; i<10; i++) {
    sample *sh;
    int pos;
    if (!channels[i].in_use)
      continue;
    sh=channels[i].sample;
    pos=channels[i].sample_position>>32;
    if (sh->sample_type & 0x10) { /* 16b sample */
      short *sample_data;
      sample_data=(short *)sh->sample_data;
      sample_data+=(pos>>1);
      mixer_soft_buffer[mixer_pos]+=((int)*sample_data*channels[i].left_volume)>>6;
      mixer_soft_buffer[mixer_pos+1]+=((int)*sample_data*channels[i].right_volume)>>6;
    } else { /* 8b sample */
      signed char *sample_data;
      sample_data=(signed char *)sh->sample_data;
      sample_data+=pos;
      /* it's -128 .. 127 and we want -32768 .. 32767, so <<8 */
      mixer_soft_buffer[mixer_pos]+=((((int)*sample_data)<<8)*channels[i].left_volume)>>6; 
      mixer_soft_buffer[mixer_pos+1]+=((((int)*sample_data)<<8)*channels[i].right_volume)>>6;
    }
    /* now, let's deal with loop or end */
    if (channels[i].increment_positif)
      channels[i].sample_position+=channels[i].increment;
    else
      channels[i].sample_position-=channels[i].increment;
    pos=channels[i].sample_position>>32;
    if(!(sh->sample_type&3)) {
      /* no loop, so test end of sample, and if yes then channel no more in use */
      if (pos>=sh->sample_length)
	channels[i].in_use=0;
    } else {
      /* let's test loop */
      if (sh->sample_type&1) {
	/* forward loop */
	if (pos>=sh->sample_loop_start+sh->sample_loop_length)
	  channels[i].sample_position=(long long)(sh->sample_loop_start)<<32;
      } else {
	/* ping pong loop */
	/* well, if >= to loop end, the reverse things, and force the sample to go to the loop end point
	 * (good ? bad ? should go back right now or what ?) same behavior in other direction
	 */
	if (channels[i].increment_positif) {
	  /* so we might be at the right end */
	  if (pos>=sh->sample_loop_start+sh->sample_loop_length) {
	    channels[i].sample_position=(long long)(sh->sample_loop_start+sh->sample_loop_length)<<32;
	    channels[i].increment_positif=0;
	  }
	} else {
	  /* might be to the left end */
	  if (pos<=sh->sample_loop_start) {
	    channels[i].sample_position=(long long)(sh->sample_loop_start)<<32;
	    channels[i].increment_positif=1;
	  }
	}
      }
    }
  }
  mixer_pos+=2;
}

int handle_tick_live(XM *module, int max_handle)
{
  int done=0;

  if (module->live_to_handle==0)
    module->live_to_handle=live_new_line(module);
  if (module->live_to_handle==0) {
    fprintf(stderr, "live_player.handle_tick_live error to_handle==0\n");
    exit(1);
  }

  while(done<max_handle && module->live_to_handle>0) {
    live_handle_one_datum(module);
    done+=8;
    module->live_to_handle-=8;
  }
  return done;
}
