/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august september 1998.
 * This is total free software.
 *-------------------------------------
 * The file contains a wav-loader and a wav-saver (and a raw-loader)
 */

#include "wav.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include "xm.h"

/* very small wav loader. There must be only 1 wav and it must not be stereo and it must be 8b or 16b */
int load_wav(sample *samp, char *name)
{
  char buf[50], *data=(char *)0;
  int len;
  int fd=open(name, O_RDONLY);
  short type;

  if (fd==-1) {
    perror("wav.load_wav 1");
    return 0;
  }

  /* wav format 1 only */
  /* first 4 bytes are "RIFF" */
  if (read(fd, buf, 4)!=4 || memcmp(buf, "RIFF", 4))
    goto error;
  /* then, a length (what length ??) */
  if (read(fd, &len, 4)!=4)
    goto error;
  /* then we have to read "WAVE" */
  if (read(fd, buf, 4)!=4 || memcmp(buf, "WAVE", 4))
    goto error;
  /* then, mikmod and sox have this while(1) that i simply copy-paste (modifying function-calls and adding the tests) */
  while(1) {
    if (read(fd, buf, 4)!=4)
      goto error;
    if (read(fd, &len, 4)!=4)
      goto error;
    if(!memcmp(buf, "fmt ", 4))
      break;
    if (lseek(fd, len, SEEK_CUR)==-1)
      goto error;
  }
  /* ok, now we read the type of wav. 1 is good, rest is not good. */
  if (read(fd, &type, 2)!=2 || type!=1)
    goto error;
  /* then we've got the nb of channels. Must be 1, otherwise it is stereo or worse, we don't handle it. */
  if (read(fd, &type, 2)!=2 || type!=1)
    goto error;
  /* then, there is : rate, average bytes/sec, block align, not used in sox, so we won't use it too */
  if (lseek(fd, 10, SEEK_CUR)==-1)
    goto error;
  /* then "FormatSpecific" says mikmod and sox uses this to know the type of sample (8, 16, 32b) */
  if (read(fd, &type, 2)!=2 || (type!=8 && type!=16))
    goto error;
  /* ok, now mikmod and sox skip len-16 bytes, so i do it too */
  if (lseek(fd, len-16, SEEK_CUR)==-1)
    goto error;
  /* now i must read "data" */
  if (read(fd, buf, 4)!=4 || memcmp(buf, "data", 4))
    goto error;
  /* now the length of the sample (if length==0, then fuck you) */
  if (read(fd, &len, 4)!=4 || len==0)
    goto error;
  /* done, now malloc, read and convert 8b samples (16b are signed ok, 8b are unsigned) */
  if (!(data=malloc(len)))
    goto error;
  if (read(fd, data, len)!=len)
    goto error1;
  if (type==8) {
    int i;
    for (i=0; i<len; i++)
      data[i]+=128;
  }
  /* now, put what has to be put in the sample structure */
  samp->sample_data=data;
  samp->sample_length=len;
  samp->sample_loop_start=0;
  samp->sample_loop_length=0;
  samp->sample_volume=0x40;
  samp->sample_fine_tune=0;
  samp->sample_relative_note_number=0;
  samp->sample_type=type==8?0:16;
  {
    int s=strlen(name);
    strcpy(samp->sample_name, &name[s-22>0?s-22:0]);
  }
  if (type==16)
    len>>=1;
  samp->max_inc_0=len;
  if (len<512)
    samp->max_inc_0=512;
  samp->inc_0=((unsigned long long)samp->max_inc_0)<<(32-9);
  samp->start_0=0;
  samp->cur_pos_0=0;
  samp->cut_in_use=0;
  /* ok, let's return (best comment i ever wrote) */
  return 1;

error1:
  free(data);
error:
  close(fd);
  return 0;
}

/* this saver is taken from sox */
int save_wav(sample *samp, char *name, int frequency)
{
  int fd;
  short dum16;
  unsigned long dum;

  if (!samp->sample_length) {
    fprintf(stderr, "No length for this sample ! what the heck d'ya wanna save ?\n");
    return 0;
  }

  if ((fd=open(name, O_WRONLY|O_TRUNC|O_CREAT, 0644))==-1) {
    perror("wav.save_wav 1");
    return 0;
  }

  /* write "RIFF" */
  if (write(fd, "RIFF", 4)!=4)
    goto error;
  /* write "waveform chunk size" (don't know what it is, directly taken from sox) */
  dum=samp->sample_length+8+16+12;	/* mikmod adds only 32, what's correct ??? */
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* write "WAVE" and "fmt " */
  if (write(fd, "WAVEfmt ", 8)!=8)
    goto error;
  /* write "fmt chunk size" (directly from sox once again) */
  dum=16;
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* write format tag (1) (sox again) (all is sox anyway) */
  dum16=1;
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* number of channels (1) */
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* the rate */  
  if (write(fd, &frequency, 4)!=4)
    goto error;
  /* average bytes/sec */
  dum=(frequency*(samp->sample_type&16?16:8)+7)/8; /* could multiply by 2, but i simply rip... */
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* block align */
  dum16=((samp->sample_type&16?16:8)+7)/8;
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* bits per sample */
  dum16=samp->sample_type&16?16:8;
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* write "data" */
  if (write(fd, "data", 4)!=4)
    goto error;
  /* sample length */
  if (write(fd, &samp->sample_length, 4)!=4)
    goto error;
  /* now write data (if 8b, must translate in unsigned then back in signed (long but no malloc) */
  if (!(samp->sample_type&16)) {
    char *data=samp->sample_data;
    int i;
    for (i=0; i<samp->sample_length; i++)
      data[i]+=128;
  }
  dum=write(fd, samp->sample_data, samp->sample_length);
  if (!(samp->sample_type&16)) {
    char *data=samp->sample_data;
    int i;
    for (i=0; i<samp->sample_length; i++)
      data[i]+=128;
  }
  if (dum!=samp->sample_length)
    goto error;

  /* done. There may be an error with close (yes it is possible, because of buffering) */
  if (close(fd)==-1) {
    perror("wav.save_wav close error");
    unlink(name);
    return 0;
  }
  return 1;

error:
  perror("wav.save_wav error");
  close(fd);
  unlink(name);
  return 0;
}

/* we gonna load it all. sample will be considered as a 8b */
int load_raw(sample *samp, char *name)
{
  int fd=open(name, O_RDONLY);
  unsigned long size;
  struct stat buf;

  if (fd==-1) {
    perror("wav.load_raw 1");
    return 0;
  }
  if (fstat(fd, &buf)==-1) {		/* if an error occurs, i kill myseld. It should not happen, the file is open */
    perror("wav.load_raw 2");
    close(fd);
    return 0; 
  }

  size=buf.st_size;

  if (samp->sample_data)
    free(samp->sample_data);	/* no test of correctness, i don't want it. */

  if (!(samp->sample_data=(char *)malloc(size))) {
    perror("wav.load_raw 3");
    close(fd);
    return 0; 
  }

  if (read(fd, samp->sample_data, size)!=size) {
    perror("wav.load_raw 4");
    free(samp->sample_data);
    close(fd);
    return 0; 
  }

  close(fd);		/* no test, things are read, don't care if a problem occurs now. */

  samp->sample_length=size;
  samp->sample_loop_start=0;
  samp->sample_loop_length=0;
  samp->sample_volume=0x40;
  samp->sample_fine_tune=0;
  samp->sample_relative_note_number=0;
  samp->sample_type=0;
  {
    int s=strlen(name);
    strcpy(samp->sample_name, &name[s-22>0?s-22:0]);
  }

  samp->max_inc_0=size;
  if (size<512)
    samp->max_inc_0=512;
  samp->inc_0=((unsigned long long)samp->max_inc_0)<<(32-9);
  samp->start_0=0;
  samp->cur_pos_0=0;
  samp->cut_in_use=0;

  return 1;
}
