/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 *-------------------------------------
 * The file contains a xm-loader and a xm-writer.
 */

#include "xm.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

#include "debug.h"
#include "mixer.h"
#include "version.h"

#include <stdio.h>

/* return 0 on error, number of bytes actually read otherwise */
int read_note(int fd, note *to)
{
  int nb=0;
  if (read(fd, &to->note, 1)!=1)
    return 0;
  nb++;
  if (!(to->note & 0x80)) {
    if (/*to->note<0 ||*/ to->note>97) /* 71 says the doc, but we find 96 notes for the instrument, bad doc */
      /* well, well, well, i forgot the cut-key which is 96... */
      /* no no no, it's 97. C0=1 not 0, fucked fuckin' doc ! */
#ifdef DEBUG
      {
	fprintf(stderr, "xm.read_note error to->note=%d which is >96\n", to->note);
#endif
      return 0;
#ifdef DEBUG
      }
#endif
    /* have to read 4 more bytes */
    if (read(fd, &to->instrument, 1)!=1 || to->instrument<=0 || (unsigned char)to->instrument>128)
      return 0;
    to->instrument--;
    if (read(fd, &to->volume, 1)!=1)
      return 0;
    if (read(fd, &to->effect_type, 1)!=1)
      return 0;
    if (read(fd, &to->effect_parameter, 1)!=1)
    return 0;
    to->note--; /* well, let's put it to its real value. And let's say -1 is for no note. */

    return 5;
  } else {
    /* have to read bytes needed */
    int readme=to->note;
    to->note=-1; /* have to put it to -1 back again. In fact C0=1 not 0 as said in the doc. Damned ! */
    to->instrument=-1;
    to->effect_type=-1;
    to->effect_parameter=0;
    to->volume=0;
    /* the order is the one given in mikmod, original doc is not very good... */
    if (readme & 1) { /* useless brace, pgcc compliance. */
      if (read(fd, &to->note, 1)!=1)
	return 0;
      else {
	to->note--;
	nb++;
      }
    }
    if (readme & 2) {
      if (read(fd, &to->instrument, 1)!=1)
	return 0;
      nb++;
      to->instrument--;
    } else
      to->instrument=255;
    if (readme & 4) { /* useless brace, pgcc compliance. */
      if (read(fd, &to->volume, 1)!=1)
	return 0;
      else
	nb++;
    }
    if (readme & 8) { /* useless brace, pgcc compliance. */
      if (read(fd, &to->effect_type, 1)!=1)
	return 0;
      else
	nb++;
    }
    if (readme & 16) { /* useless brace, pgcc compliance. */
      if (read(fd, &to->effect_parameter, 1)!=1)
	return 0;
      else
	nb++;
    }

    /* here we assume effect_type=0 if a parameter is given (sanx to Ripper for walker.xm) */
    if (to->effect_parameter && to->effect_type==-1)
      to->effect_type=0;

    return nb;
  }
}

/* load the file "name" in the "xm" structure, return 0 on error, 1 otherwise */
int load_xm(char *name, XM *xm)
{
  int fd, i, dummy;
  char dum[100]; /* it won't buffer-overflow, who you think i am, huh ? */

  memset(dum, 0, sizeof(dum));
  memset(xm, 0, sizeof(XM));

  /* is the file readable ? */
  if (access(name, R_OK)==-1)
    return 0;
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm file readable ok\n");
#endif

  /* ok, let's load it */
  if ((fd=open(name, O_RDONLY))==-1)
    return 0;
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm file open ok\n");
#endif

  /* first, check the "Extended Module: " string beware of the 'M' in "Module", doc says "module"... */
  if (read(fd, dum, 17)!=17)
    goto error0;
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm first read ok\ndum = %s\n", dum);
#endif
  if (strncmp(dum, "Extended Module: ", 17))
    goto error0;
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm \"extended module\" check ok\n");
#endif

  /* ok, read the module name */
  if (read(fd, xm->module_name, 20)!=20)
    goto error0;
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm module name=%s\n", xm->module_name);
#endif
  /* let's forget the last space-bytes */
  for(i=19; i>=0; i--)
    if (xm->module_name[i]!=' ')
      break;
    else
      xm->module_name[i]=0;

  /* must read the byte 0x1a */
  if (read(fd, dum, 1)!=1 || dum[0] != 0x1a)
    goto error0;

  /* read the tracker-name (if not "xsoundtrack X.Y.Z", format the hdd) (joke) */
  if (read(fd, xm->tracker_name, 20)!=20)
    goto error0;
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm tracker name=%s\n", xm->tracker_name);
#endif

  /* ok, read the rest of the header */
  if (read(fd, &xm->version, 2)!=2) {
    close(fd);
    return 0;
  }
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm version=%x\n", xm->version);
#endif
  if (read(fd, &xm->header_size, 4)!=4) {
    close(fd);
    return 0;
  }
  if (read(fd, &xm->song_length, 2)!=2 || xm->song_length>256 || xm->song_length<0) {
    close(fd);
    return 0;
  }
  if (xm->song_length==0) {
    fprintf(stderr, "xm.load_xm error : song length=0. Contact me to report, should not be possible...\n");
    return 0;
  }
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm song_length=%d\n", xm->song_length);
#endif
  if (read(fd, &xm->restart, 2)!=2 || xm->restart<0 || xm->restart>=xm->song_length) {
    close(fd);
    return 0;
  }
  if (read(fd, &xm->nb_channels, 2)!=2 || xm->nb_channels<2 || xm->nb_channels>32 || (xm->nb_channels & 1)) {
    close(fd);
    return 0;
  }
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm nb_channels=%d\n", xm->nb_channels);
#endif
  if (read(fd, &xm->nb_patterns, 2)!=2 || xm->nb_patterns<0 || xm->nb_patterns>256) {
    close(fd);
    return 0;
  }
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm nb_patterns=%d\n", xm->nb_patterns);
#endif
  if (read(fd, &xm->nb_instruments, 2)!=2 || xm->nb_instruments<0 || xm->nb_instruments>128) {
    close(fd);
    return 0;
  }
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm nb_instruments=%d\n", xm->nb_instruments);
#endif
  if (read(fd, &xm->flags, 2)!=2 || xm->flags & ~1) {
    close(fd);
    return 0;
  }
  if (read(fd, &xm->default_tempo, 2)!=2) {
    close(fd);
    return 0;
  }
  if (read(fd, &xm->default_bpm, 2)!=2) {
    close(fd);
    return 0;
  }
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm default_tempo=%d default_bpm=%d\n", xm->default_tempo, xm->default_bpm);
#endif
  if (read(fd, xm->patterns_order, 256)!=256) {
    close(fd);
    return 0;
  }

  /* ok, we finished the header (and so the fixed-datas), now come the patterns (and so some mallocs). */
  memset(xm->patterns, 0, sizeof(xm->patterns));

  for (i=0; i<xm->nb_patterns; i++) {
    int nb_bytes_read=0;
    int line;
    note *cur=(note *)0; /* GCC complains that this var might be used uninitialized !!! */
    /* don't know what to do with the pattern header length, so just verify it makes 9 bytes */
    if (read(fd, &dummy, 4)!=4 || dummy!=9)
      goto error;
    dummy=0;
    /* then we read "packing type", a byte that must be 0 */
    if (read(fd, &dummy, 1)!=1 || dummy)
      goto error;
    /* then, number of rows */
    if (read(fd, &xm->patterns_nb_rows[i], 2)!=2)
      goto error;
#ifdef DEBUG
    fprintf(stderr, "xm.load_xm patterns_nb_rows[%d]=%d\n", i, xm->patterns_nb_rows[i]);
#endif
    /* then, size of the data to read (don't really need it, we've got the rows, and the nb of channels
     * maybe later for verification)
     */
    /* well, seems that 0 len is possible... fuckin' format ! */
    dummy=0;
    if (read(fd, &dummy, 2)!=2)
      goto error;
#ifdef DEBUG
    fprintf(stderr, "xm.load_xm size of data to read=%d\n", dummy);
#endif
    /* now, alloc some memory (even if no data to read, problem with a xm (elektron.xm)) */
    /* if nothing is to be read, well, there's a problem anyway so don't care what will be in the pattern */
    if (xm->patterns_nb_rows[i]) {
      if (!(cur=xm->patterns[i]=(note *)malloc(sizeof(note)*xm->nb_channels*xm->patterns_nb_rows[i])))
	goto error;
      memset(cur, 0, sizeof(note)*xm->nb_channels*xm->patterns_nb_rows[i]);
      if (!dummy) {
	for (line=0; line<xm->patterns_nb_rows[i]; line++) {
	  int cur_note;
	  for (cur_note=0; cur_note<xm->nb_channels; cur_note++) {
	    note *n=&cur[line+cur_note*xm->patterns_nb_rows[i]];
	    n->instrument=-1;
	    n->note=-1;
	    n->effect_type=-1;
	  }
	}
      }
    }
    if (dummy) {
      /* read the notes. They are put line by line (like what you see in ft2) (and like in mikmod) */
      for (line=0; line<xm->patterns_nb_rows[i]; line++) {
	int cur_note;
	for (cur_note=0; cur_note<xm->nb_channels; cur_note++) {
	  int dum=read_note(fd, &cur[line+cur_note*xm->patterns_nb_rows[i]]);
	  if (!dum)
	    goto error;
	  nb_bytes_read+=dum;
	}
      }
    }
    /* if we did not read the correct number of bytes, then error */
    if (nb_bytes_read!=dummy)
#ifdef DEBUG
      {
	fprintf(stderr, "xm.load_xm error nb_bytes_read=%d dummy=%d, which is !=\n", nb_bytes_read, dummy);
#endif
      goto error;
#ifdef DEBUG
      }
#endif
  }

  /* ok, patterns are read, now the instruments */
  memset(xm->instruments, 0, sizeof(xm->instruments));
  for (i=0; i<xm->nb_instruments; i++) {
    int j, inst_size;

    /* read the instrument size. What is it ? number of bytes taken in the file ? don't know what to do with it,
     * maybe sanity check...
     */
    /* well, seems it's usefull, for the "!    +241      2   (word) Reserved" stuff. Sanx to mikmod 3.0.3
     * (2.14 was right, but i didn't understand.)
     */
    if (read(fd, &inst_size, 4)!=4)
      goto error1;
#ifdef DEBUG
  fprintf(stderr, "inst_size=%d\n", inst_size);
#endif
    inst_size-=4;
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm instrument size=%d ########\n", inst_size);
#endif
    
    /* malloc the instrument descriptor */
    if (!(xm->instruments[i]=(instrument *)malloc(sizeof(instrument))))
      goto error1;

    memset(xm->instruments[i], 0, sizeof(instrument));

    /* read the name of the instrument */
    if (read(fd, &xm->instruments[i]->name, 22)!=22)
      return 0;
    inst_size-=22;
    /* forget the last ' ' (line-coding, new style) */
    { int j; for (j=21; j>=0; j--) if (xm->instruments[i]->name[j]!=' ') break; else xm->instruments[i]->name[j]=0; }
#ifdef DEBUG
    fprintf(stderr, "xm.load_xm name = %s\n", xm->instruments[i]->name);
#endif
    /* read instrument type (must be 0 says the doc, but in fact, well, you know...) */
    dummy=0;
    if (read(fd, &dummy, 1)!=1) /*  || dummy) : don't test it, it means nothing */
      goto error1;
    inst_size--;
#ifdef DEBUG
    fprintf(stderr, "xm.load_xm dummy = %d\n", dummy);
#endif
    /* read number of samples */
    if (read(fd, &xm->instruments[i]->nb_samples, 2)!=2)
      goto error1;
    inst_size-=2;
#ifdef DEBUG
    fprintf(stderr, "xm.load_xm nb of samples=%d\n", xm->instruments[i]->nb_samples);
#endif
    if (!(xm->instruments[i]->nb_samples)) {
#ifdef DEBUG
      fprintf(stderr, "now inst_size=%d\n", inst_size);
#endif
      while(inst_size) {
	/* well, just for me to see what the value is... no need otherwise, and should be 4 anyway */
	if (inst_size>=4) {
	  if (read(fd, &dummy, 4)!=4)
	    goto error1;
	  inst_size-=4;
#ifdef DEBUG
	  fprintf(stderr, "xm.load_xm dummy in the quitting sequence is = %d\n", dummy);
#endif	  
	} else {
	  if (read(fd, &dummy, 1)!=1)
	    goto error1;
	  inst_size--;
	}
      }
      continue;
    }

    /* read the header it-self */
    /* sample-header size (what to do with it ? sample header size is 40 bytes, so ? */
    if (read(fd, &dummy, 4)!=4)
      goto error1;
    inst_size-=4;

#ifdef DEBUG
    fprintf(stderr, "xm.load_xm sample header size = %d\n", dummy);
#endif
    /* sample number for all notes */
    if (read(fd, &xm->instruments[i]->sample_number[0], 96)!=96)
      goto error1;
    inst_size-=96;
    /* points for volume envelop */
    if (read(fd, &xm->instruments[i]->volume_envelop[0][0], 48)!=48)
      goto error1;
    inst_size-=48;
    /* points for panning envelop */
    if (read(fd, &xm->instruments[i]->panning_envelop[0][0], 48)!=48)
      goto error1;
    inst_size-=48;
    /* number of volume points */
    if (read(fd, &xm->instruments[i]->nb_volume_points, 1)!=1)
      goto error1;
    inst_size--;
    /* number of panning points */
    if (read(fd, &xm->instruments[i]->nb_panning_points, 1)!=1)
      goto error1;
    inst_size--;
#ifdef DEBUG
    fprintf(stderr, "xm.load_xm nb vol points = %d nb panning points = %d\n",
	    xm->instruments[i]->nb_volume_points, xm->instruments[i]->nb_panning_points);
    {
      int al;
      short *p=(short *)&xm->instruments[i]->volume_envelop[0];
      for (al=0; al<xm->instruments[i]->nb_volume_points; al++)
	fprintf(stderr, "xm.load_xm vol points[%d] = %d %d\n", al, p[al*2], p[al*2+1]);
    }
    {
      int al;
      short *p=(short *)&xm->instruments[i]->panning_envelop[0];
      for (al=0; al<xm->instruments[i]->nb_volume_points; al++)
	fprintf(stderr, "xm.load_xm pan points[%d] = %d %d\n", al, p[al*2], p[al*2+1]);
    }
    fprintf(stderr, "press a key ---");
    getchar();
#endif
    /* volume sustain point */
    if (read(fd, &xm->instruments[i]->volume_sustain_point, 1)!=1)
      goto error1;
    inst_size--;
    /* volume loop start */
    if (read(fd, &xm->instruments[i]->volume_loop_start, 1)!=1)
      goto error1;
    inst_size--;
    /* volume loop end */
    if (read(fd, &xm->instruments[i]->volume_loop_end, 1)!=1)
      goto error1;
    inst_size--;
#ifdef DEBUG
    fprintf(stderr, "vol sus = %d loop star = %d loop end = %d\n",
	    xm->instruments[i]->volume_sustain_point, xm->instruments[i]->volume_loop_start,
	    xm->instruments[i]->volume_loop_end);
#endif
    /* panning sustain point */
    if (read(fd, &xm->instruments[i]->panning_sustain_point, 1)!=1)
      goto error1;
    inst_size--;
    /* panning loop start */
    if (read(fd, &xm->instruments[i]->panning_loop_start, 1)!=1)
      goto error1;
    inst_size--;
    /* panning loop end */
    if (read(fd, &xm->instruments[i]->panning_loop_end, 1)!=1)
      goto error1;
    inst_size--;
    /* volume type */
    if (read(fd, &xm->instruments[i]->volume_type, 1)!=1)
      goto error1;
    inst_size--;
    /* panning type */
    if (read(fd, &xm->instruments[i]->panning_type, 1)!=1)
      goto error1;
    inst_size--;
    /* vibrato type */
    if (read(fd, &xm->instruments[i]->vibrato_type, 1)!=1)
      goto error1;
    inst_size--;
    /* vibrato sweep */
    if (read(fd, &xm->instruments[i]->vibrato_sweep, 1)!=1)
      goto error1;
    inst_size--;
    /* vibrato depth */
    if (read(fd, &xm->instruments[i]->vibrato_depth, 1)!=1)
      goto error1;
    inst_size--;
    /* vibrato rate */
    if (read(fd, &xm->instruments[i]->vibrato_rate, 1)!=1)
      goto error1;
    inst_size--;
#ifdef DEBUG
    fprintf(stderr, "xm.load_xm vibrato type=%d sweep=%d depth=%d rate=%d\n", xm->instruments[i]->vibrato_type,
	    xm->instruments[i]->vibrato_sweep, xm->instruments[i]->vibrato_depth, xm->instruments[i]->vibrato_rate);
#endif
    /* volume fadeout */
    if (read(fd, &xm->instruments[i]->volume_fadeout, 2)!=2)
      goto error1;
    inst_size-=2;
    /* 22 reserved bytes (why 22 ???) */
    for (; inst_size>0; inst_size--)
      if (read(fd, &dummy, 1)!=1)
	goto error1;
    /* let's read the samples */
    {
      /* first, we find the headers (at the begining, i though samples were stored one after the other,
       * but we first have the headers, then the samples them-self, argh ! did not understand the doc...
       */
      for (j=0; j<xm->instruments[i]->nb_samples; j++) {
	sample *sh=&(xm->instruments[i]->sample_headers[j]);
#ifdef DEBUG
	fprintf(stderr, "xm.load_xm xm->instrument[%d]->nb_sample = %d sh = %p j = %d\n",
		i, xm->instruments[i]->nb_samples, sh, j);
#endif
	/* the header */
	/* the sample length (in bytes ? or in sample-type (ie number of notes or bytes ? what with 16b samples ?
	 * Well, seems to be number of bytes (mikmod does not seem to read the data... not in that function)
	 */
#ifdef DEBUG
	fprintf(stderr, "xm.load_xm pass 0\n");
#endif
	if (read(fd, &sh->sample_length, 4)!=4) {
	  perror("read");
	  goto error1;
	}
#ifdef DEBUG
	fprintf(stderr, "xm.load_xm sample length = %d\n", sh->sample_length);
#endif
	/* loop start */
	if (read(fd, &sh->sample_loop_start, 4)!=4)
	  goto error1;
	/* loop length (stupid. Why not simply loop end ?) */
	if (read(fd, &sh->sample_loop_length, 4)!=4)
	  goto error1;
	/* volume */
	if (read(fd, &sh->sample_volume, 1)!=1)
	  goto error1;
	/* fine tune */
	if (read(fd, &sh->sample_fine_tune, 1)!=1)
	  goto error1;
	/* type */
	if (read(fd, &sh->sample_type, 1)!=1)
	  goto error1;
	/* panning */
	if (read(fd, &sh->sample_panning, 1)!=1)
	  goto error1;
	/* relative note number */
	if (read(fd, &sh->sample_relative_note_number, 1)!=1)
	  goto error1;
#ifdef DEBUG
	fprintf(stderr, "xm.load_xm relative tone number = %d\n", sh->sample_relative_note_number);
#endif
	/* reserved byte */
	if (read(fd, &dummy, 1)!=1)
	  goto error1;
	/* sample name */
	if (read(fd, &sh->sample_name[0], 22)!=22)
	  goto error1;
	/* forget the last ' ' (line-coding, new style) */
	{ int j; for (j=21; j>=0; j--) if (sh->sample_name[j]!=' ') break; else sh->sample_name[j]=0; }
#ifdef DEBUG
	fprintf(stderr, "xm.load_xm sample name = %s\n", sh->sample_name);
#endif
	/* now, data. So, doc says delta values. Have to make difference between 8 and 16b */
	/* first of all, malloc space */
	if (sh->sample_length)
	  if (!(sh->sample_data=(char *)malloc(sh->sample_length)))
	    goto error1;
      }

      for (j=0; j<xm->instruments[i]->nb_samples; j++) {
	sample *sh=&(xm->instruments[i]->sample_headers[j]);
	if (read(fd, sh->sample_data, sh->sample_length)!=sh->sample_length)
	  goto error1;
	if (sh->sample_type & 0x10) {
	  /* 16b */
	  int l;
	  short old=0, new;
	  for (l=0; l<(sh->sample_length>>1); l++) {
	    new=*(short *)(((short *)(sh->sample_data))+l)+old;
	    *(short *)((short *)sh->sample_data+l)=new;
	    old=new;
	  }
	} else {
	  int l;
	  /* 8b */
	  signed char old=0, new;
	  for (l=0; l<sh->sample_length; l++) {
	    new=*(signed char *)(((signed char *)(sh->sample_data))+l)+old;
	    *(signed char *)(((signed char *)(sh->sample_data))+l)=new;
	    old=new;
	  }
	}
      }
    }
  }

  /* done. Woo, hard coding ! */
  close(fd);

  /* ok, now we can play some patterns which might have a 0-length (have some of those in my hd),
   * so, we verify, and if so, we create a 0 filled pattern of length 64, so we won't have any troubles (hope so)
   */
  {
    int i;
    for(i=0; i<xm->song_length; i++)
      if (!xm->patterns_nb_rows[xm->patterns_order[i]]) { /* if length=0, then no pattern there, so alloc one */
	xm->patterns[xm->patterns_order[i]]=(note *)malloc(64*xm->nb_channels*sizeof(note));
	if (!xm->patterns[xm->patterns_order[i]]) {
	  fprintf(stderr, "xm.load_xm : warning malloc failed\n");
	  goto error1;
	}
	xm->patterns_nb_rows[xm->patterns_order[i]]=64;
	memset(xm->patterns[xm->patterns_order[i]], 0xff, 64*xm->nb_channels*sizeof(note));
	{
	  int j;
	  for (j=0; j<64*xm->nb_channels; j++)
	    xm->patterns[xm->patterns_order[i]][j].volume=0,
	    xm->patterns[xm->patterns_order[i]][j].effect_parameter=0;
	}
      }
  }
  return 1;

error1:
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm error1 error\n");
#endif
  i=0;
  while(i<128 && xm->instruments[i]) {
    if (xm->instruments[i]->sample_headers) {
      int j;
      for (j=0; j<xm->instruments[i]->nb_samples; j++)
	if (xm->instruments[i]->sample_headers[j].sample_data)
	  free(xm->instruments[i]->sample_headers[j].sample_data);
    }
    
    free(xm->instruments[i]);
    i++;
  }
error:
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm error error\n");
#endif
  i=0;
  while(i<256 && xm->patterns[i])
    free(xm->patterns[i++]);
error0:
#ifdef DEBUG
  fprintf(stderr, "xm.load_xm error0 error\n");
#endif
  close(fd);
  return 0;
}

/* return 0 on error 1 in other cases */
int write_xm_string(int fd, char *c, int n)
{
  int i;
  for (i=n-1; i>=0; i--)
    if (c[i])
      break;
    else
      c[i]=' ';
  if (write(fd, c, n)!=n)
    return 0;
  return 1;
}

/* return nb of bytes put in file */
int write_note(int fd, note n)
{
  char c;
  int nb=1;

  /* if all is there, write the 5 bytes */
  /* don't write effect type if = 0 */
  if (n.note!=255 && n.instrument!=-1 && (n.effect_type!=-1 &&n.effect_type) && n.effect_parameter!=0 && n.volume!=0) {
    c=n.note+1;
    if (write(fd, &c, 1)!=1)
      return 0;
    c=n.instrument+1;
    if (write(fd, &c, 1)!=1)
      return 0;
    c=n.volume;
    if (write(fd, &c, 1)!=1)
      return 0;
    c=n.effect_type;
    if (write(fd, &c, 1)!=1)
      return 0;
    c=n.effect_parameter;
    if (write(fd, &c, 1)!=1)
      return 0;

    return 5;
  }

  /* only some of it is there */
  c=0x80;
  if (n.note!=255)
    c|=1, nb++;
  if (n.instrument!=-1)
    c|=2, nb++;
  if (n.volume!=0)
    c|=4, nb++;
  if (n.effect_type!=-1 && n.effect_type)
    c|=8, nb++;
  if (n.effect_parameter!=0)
    c|=16, nb++;
  if (write(fd, &c, 1)!=1)
    return 0;
  if (n.note!=255) {
    c=n.note+1;
    if (write(fd, &c, 1)!=1)
      return 0;
  }
  if (n.instrument!=-1) {
    c=n.instrument+1;
    if (write(fd, &c, 1)!=1)
      return 0;
  }
  if (n.volume!=0) {
    c=n.volume;
    if (write(fd, &c, 1)!=1)
      return 0;
  }
  if (n.effect_type!=-1 && n.effect_type) {
    c=n.effect_type;
    if (write(fd, &c, 1)!=1)
      return 0;
  }
  if (n.effect_parameter!=-0) {
    c=n.effect_parameter;
    if (write(fd, &c, 1)!=1)
      return 0;
  }

  return nb;
}

int save_xm(XM *xm, char *name)
{
  int fd;
  char c;
  short dum16;
  int dum32, i, j;

  if (!name) {
    fprintf(stderr, "xm.save_xm error : the file name is void, what do you do ? (or maybe there is no more RAM)\n");
    return 0;
  }

  if ((fd=open(name, O_WRONLY|O_TRUNC|O_CREAT, 0644))==-1) {
    perror("xm.save_xm 1");
    return 0;
  }

  /* write "Extended Module: " */
  #define EM		"Extended Module: "
  if (write(fd, EM, strlen(EM))!=strlen(EM))
    goto error;
  /* write the module name */
  if (!write_xm_string(fd, xm->module_name, 20))
    goto error;
  /* write the byte 0x1a */
  c=0x1a;
  if (write(fd, &c, 1)!=1)
    goto error;
  /* write the tracker name */
  {
    char nom[21];
    int i;
    nom[20]=0;
    memset(nom, ' ', 20);
    for (i=0; i<strlen(TN); i++)
      nom[i]=TN[i];
    if (write(fd, nom, 20)!=20)
      goto error;
  }
  /* write the version (it's 1.04, it's ft2 version) */
  dum16=0x0104;
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* write the header size (276 bytes) */
  dum32=276;
  if (write(fd, &dum32, 4)!=4)
    goto error;
  /* write the song length */
  if (write(fd, &xm->song_length, 2)!=2)
    goto error;
  /* restart position */
  if (write(fd, &xm->restart, 2)!=2)
    goto error;
  /* nb of channels */
  if (write(fd, &xm->nb_channels, 2)!=2)
    goto error;
  /* nb of patterns (has to be calculated) */
  for (dum32=0; dum32<256; dum32++)
    if (xm->patterns_nb_rows[dum32])
      xm->nb_patterns=dum32+1;
  if (xm->nb_patterns>255)
    xm->nb_patterns=255;
  /* it is at least 1, impossible to have 0 */
  if (write(fd, &xm->nb_patterns, 2)!=2)
    goto error;
  /* nb of instruments (once again has to be calculated) */
  xm->nb_instruments=0;
  for (dum32=0; dum32<128; dum32++)
    if (xm->instruments[dum32])
      xm->nb_instruments=(short)dum32+1;
  if (write(fd, &xm->nb_instruments, 2)!=2)
    goto error;
  /* flags (bit 0 to say if linear or amiga freq. table) */
  if (write(fd, &xm->flags, 2)!=2)
    goto error;
  /* default tempo/bpm */
  if (write(fd, &xm->default_tempo, 2)!=2)
    goto error;
  if (write(fd, &xm->default_bpm, 2)!=2)
    goto error;
  /* pattern order */
  if (write(fd, xm->patterns_order, 256)!=256)
    goto error;
  /* now, the patterns */
  for (i=0; i<xm->nb_patterns; i++) {
    short dat_size=0;
    off_t cur_pos;
    int line;
    note *cur;
    /* write the pattern header */
    /* first, its length (9 bytes) */
    dum32=9;
    if (write(fd, &dum32, 4)!=4)
      goto error;
    /* packing type (0 in fact, there's only 1 packing method anyway) */
    c=0;
    if (write(fd, &c, 1)!=1)
      goto error;
    /* nb of rows */
    if (write(fd, &xm->patterns_nb_rows[i], 2)!=2)
      goto error;
    /* now, we must put the size of the data (we don't know that for now, so we take the current position,
     * then we write the data (an count how many bytes were writen (wroten ?)) then we go back here to write
     * their size
     */
    if ((cur_pos=lseek(fd, 0, SEEK_CUR))==-1)
      goto error;
    /* we write 0 because we may have 0 length pattern */
    dum16=0;
    if (write(fd, &dum16, 2)!=2)
      goto error; 
    cur=xm->patterns[i];
    for (line=0; line<xm->patterns_nb_rows[i]; line++) {
      int cur_note;
      for (cur_note=0; cur_note<xm->nb_channels; cur_note++) {
	int dum=write_note(fd, cur[line+cur_note*xm->patterns_nb_rows[i]]);
	if (!dum)
	  goto error;
	dat_size+=dum;
      }
    }
    if (lseek(fd, cur_pos, SEEK_SET)==-1)
      goto error;
    if (write(fd, &dat_size, 2)!=2)
      goto error;
    if (lseek(fd, 0, SEEK_END)==-1)
      goto error;
  }
  /* now, the instruments */
  for (i=0; i<xm->nb_instruments; i++) {
    instrument *inst=xm->instruments[i];
    /* we must write the instrument header size. It is 263 if there is one or more samples, 33 otherwise */
    /* well, hope it's correct... */
    /* if instrument==(instrument *)0, then we have to put a void instrument, otherwise, next loading will have
     * strange results. (instrument 2 would be saved as 1, bad...)
     * if no sample is here, that's the same
     */
    if (inst) {
      int i;
      inst->nb_samples=0;
      for (i=0; i<16; i++)
	if (inst->sample_headers[i].sample_length || strlen(inst->sample_headers[i].sample_name))
	  inst->nb_samples=i+1;
    }
    /* write the length of the header */
    if (!inst || inst->nb_samples==0)
      dum32=33;
    else
      dum32=263;
    if (write(fd, &dum32, 4)!=4)
      goto error;
    /* the instrument name, if there is one */
    if (inst) {
      if (!write_xm_string(fd, inst->name, 22))
	goto error;
    } else
      if (!write_xm_string(fd, "                      ", 22))
	goto error;
    /* then write 0, even if ft2 seems to write dummy (well, hope it's dummy) */
    c=0;
    if (write(fd, &c, 1)!=1)
      goto error;
    /* write nb of samples */
    dum16=inst?inst->nb_samples:0;
    if (write(fd, &dum16, 2)!=2)
      goto error;
    /* write sample header size (40 bytes) */
    dum32=40;
    if (write(fd, &dum32, 4)!=4)
      goto error;
    /* if no sample, stop here */
    if (!inst || inst->nb_samples==0)
      continue;
    /* There is an instrument and at least one sample. */
    /* write sample number for all notes */
    if (write(fd, &inst->sample_number[0], 96)!=96)
      goto error;
    /* points for volume envelope */
    if (write(fd, &inst->volume_envelop[0][0], 48)!=48)
      goto error;
    /* points for panning envelope */
    if (write(fd, &inst->panning_envelop[0][0], 48)!=48)
      goto error;
    /* nb of volume points */
    if (write(fd, &inst->nb_volume_points, 1)!=1)
      goto error;
    /* nb of panning points */
    if (write(fd, &inst->nb_panning_points, 1)!=1)
      goto error;
    /* volume sustain point */
    if (write(fd, &inst->volume_sustain_point, 1)!=1)
      goto error;
    /* volume loop start */
    if (write(fd, &inst->volume_loop_start, 1)!=1)
      goto error;
    /* volume loop end */
    if (write(fd, &inst->volume_loop_end, 1)!=1)
      goto error;
    /* panning sustain point */
    if (write(fd, &inst->panning_sustain_point, 1)!=1)
      goto error;
    /* panning loop start */
    if (write(fd, &inst->panning_loop_start, 1)!=1)
      goto error;
    /* panning loop end */
    if (write(fd, &inst->panning_loop_end, 1)!=1)
      goto error;
    /* volume type */
    if (write(fd, &inst->volume_type, 1)!=1)
      goto error;
    /* panning type */
    if (write(fd, &inst->panning_type, 1)!=1)
      goto error;
    /* vibrato type */
    if (write(fd, &inst->vibrato_type, 1)!=1)
      goto error;
    /* vibrato sweep */
    if (write(fd, &inst->vibrato_sweep, 1)!=1)
      goto error;
    /* vibrato depth */
    if (write(fd, &inst->vibrato_depth, 1)!=1)
      goto error;
    /* vibrato rate */
    if (write(fd, &inst->vibrato_rate, 1)!=1)
      goto error;
    /* volume fadeout */
    if (write(fd, &inst->volume_fadeout, 2)!=2)
      goto error;
    /* 22 reserved bytes, don't ask me why, that's just the way it is... */
    {
      char dum[22];
      memset(dum, 0, 22);
      if (write(fd, &dum, 22)!=22)
	goto error;
    }
    /* now come the headers of the samples */
    for (j=0; j<inst->nb_samples; j++) {
      sample *sh=&inst->sample_headers[j];
      /* write the size of the sample */
      if (write(fd, &sh->sample_length, 4)!=4)
	goto error;
      /* sample loop start */
      if (write(fd, &sh->sample_loop_start, 4)!=4)
	goto error;
      /* sample loop end */
      if (write(fd, &sh->sample_loop_length, 4)!=4)
	goto error;
      /* volume */
      if (write(fd, &sh->sample_volume, 1)!=1)
	goto error;
      /* fine tune */
      if (write(fd, &sh->sample_fine_tune, 1)!=1)
	goto error;
      /* type */
      if (write(fd, &sh->sample_type, 1)!=1)
	goto error;
      /* panning */
      if (write(fd, &sh->sample_panning, 1)!=1)
	goto error;
      /* relative note number */
      if (write(fd, &sh->sample_relative_note_number, 1)!=1)
	goto error;
      /* reserved byte (we gonna put 0 then) */
      c=0;
      if (write(fd, &c, 1)!=1)
	goto error;
      /* sample name */
      if (!write_xm_string(fd, &sh->sample_name[0], 22))
	goto error;
      /* that's all. Data come after */
    }
    /* now put the samples data */
    for (j=0; j<inst->nb_samples; j++) {
      sample *sh=&inst->sample_headers[j];
      if (!sh->sample_length)
	continue;
      /* let's code it */
      if (sh->sample_type & 0x10) {
	/* 16b */
	int l;
	short old=*(short *)(((short *)(sh->sample_data))), new;
	for (l=1; l<(sh->sample_length>>1); l++) {
	  new=*(short *)(((short *)(sh->sample_data))+l);
	  *(short *)((short *)sh->sample_data+l)-=old;
	  old=new;
	}
      } else {
	int l;
	/* 8b */
	signed char old=*(signed char *)(((signed char *)(sh->sample_data))), new;
	for (l=1; l<sh->sample_length; l++) {
	  new=*(signed char *)(((signed char *)(sh->sample_data))+l);
	  *(signed char *)(((signed char *)(sh->sample_data))+l)-=old;
	  old=new;
	}
      }
      /* now save it */
      if (write(fd, sh->sample_data, sh->sample_length)!=sh->sample_length)
	goto error;
      /* now, decode it */
      if (sh->sample_type & 0x10) {
	/* 16b */
	int l;
	short old=0, new;
	for (l=0; l<(sh->sample_length>>1); l++) {
	  new=*(short *)(((short *)(sh->sample_data))+l)+old;
	  *(short *)((short *)sh->sample_data+l)=new;
	  old=new;
	}
      } else {
	int l;
	/* 8b */
	signed char old=0, new;
	for (l=0; l<sh->sample_length; l++) {
	  new=*(signed char *)(((signed char *)(sh->sample_data))+l)+old;
	  *(signed char *)(((signed char *)(sh->sample_data))+l)=new;
	  old=new;
	}
      }
    }
  }

  if (close(fd)==-1) {
    perror("xm.save_xm save error");
    unlink(name);
    return 0;
  }

  return 1;

error:
  perror("xm.save_xm error");
  close(fd);
  unlink(name);
  return 0;
}

/* we dump in 16b, at 44100Hz, mono. But the mixer creates stereo, so we have to remix things.
 * Later, should make an option to choose output resolution, here and in the mixer.
 * Well, later. It's gonna be very bug-generator, and there are enough in this prog.
 */
int dump_wav(XM *xm, char *name)
{
  extern int mixer_soft_buffer[1024*2/4];
  extern char mixer_buffer_soundcard[1024];
  int *from;
  short *to;
  int i, val, inc;
  int fd;
  int dum;
  short dum16;
  unsigned long dump_size=0;

  if (!name) {
    fprintf(stderr, "xm.dump_wav error : the file name is void, what do you do ? (or maybe there is no more RAM)\n");
    return 0;
  }

  if ((fd=open(name, O_WRONLY|O_TRUNC|O_CREAT, 0644))==-1) {
    perror("xm.dump_wav 1");
    return 0;
  }

  /* let's write the header */
  /* write "RIFF" */
  if (write(fd, "RIFF", 4)!=4)
    goto error;
  /* write "waveform chunk size" (we don't know the length for now, write it back later */
  dum=0;
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* write "WAVE" and "fmt " */
  if (write(fd, "WAVEfmt ", 8)!=8)
    goto error;
  /* write "fmt chunk size" (directly from sox once again) */
  dum=16;
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* write format tag (1) (sox again) (all is sox anyway) */
  dum16=1;
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* number of channels (1) */
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* the rate */
  dum=44100;
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* average bytes/sec */
  dum=(44100*16+7)/8;
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* block align */
  dum16=(16+7)/8;
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* bits per sample */
  dum16=16;
  if (write(fd, &dum16, 2)!=2)
    goto error;
  /* write "data" */
  if (write(fd, "data", 4)!=4)
    goto error;
  /* sample length (unknown for now, write it back later) */
  dum=0;
  if (write(fd, &dum, 4)!=4)
    goto error;

  /* we dump from 0 of course */
  xm->mix=MODULE_PLAY;
  xm->current_pattern=0;
  xm->current_line=-1;
  xm->current_tempo=xm->default_tempo;
  xm->speed=32768;
  xm->bpm=xm->default_bpm;
  xm->global_volume=64;
  xm->do_we_dump=1;
  while(xm->do_we_dump) {
    mix(xm);
    from=mixer_soft_buffer;
    to=(short *)mixer_buffer_soundcard;
    for (i=0; i<256; i++, to++, from+=2) {
      val=((*from)+(*(from+1)))>>1;
      if (val<-32768)
	val=-32768;
      if (val>32767)
	val=32767;
      *to=(short)val;
    }
    if ((inc=write(fd, mixer_buffer_soundcard, xm->dumped))!=xm->dumped)
      goto error;
    dump_size+=inc;
  }

  /* now, we know the length, let's go write it */
  /* pos 4 - chunk size */
  if (lseek(fd, 4, SEEK_SET)!=4)
    goto error;
  dum=dump_size+8+16+12;
  if (write(fd, &dum, 4)!=4)
    goto error;
  /* pos 40 - sample size */
  if (lseek(fd, 40, SEEK_SET)!=40)
    goto error;
  if (write(fd, &dump_size, 4)!=4)
    goto error;

  if (close(fd)==-1)
    goto error;

  xm->mix=0;
  xm->current_pattern=0;
  xm->current_line=-1;
  xm->current_tempo=xm->default_tempo;
  xm->speed=32768;
  xm->bpm=xm->default_bpm;
  xm->global_volume=64;

  return 1;

error:
  perror("xm.dump_wav error");
  close(fd);
  unlink(name);

  xm->mix=0;
  xm->current_pattern=0;
  xm->current_line=-1;
  xm->current_tempo=xm->default_tempo;
  xm->speed=32768;
  xm->bpm=xm->default_bpm;
  xm->global_volume=64;

  return 0;
}

/* it simply frees a xm. */
/* we don't test if the frees are done correctly, that's bad. */
void free_xm(XM *xm)
{
  int i;
  /* first, run through the instrument, and free it if there is one. Don't forget to free the samples */
  for (i=0; i<128; i++) {
    int j;
    instrument *inst=xm->instruments[i];
    if (!inst)
      continue;
    for (j=0; j<16; j++) {
      sample *samp=&inst->sample_headers[j];
      if (samp->sample_data)
	free(samp->sample_data);
    }
    free(inst);
  }
  /* then, free the patterns */
  for (i=0; i<256; i++) {
    note *n=xm->patterns[i];
    if (!n)
      continue;
    free(n);
  }
  /* ok, now free it */
  free(xm);
  /* done */
}
