/*
 * This file is part of xsoundtrack - a sound tracker for Linux.
 * Sed - july august 1998.
 * This is total free software.
 */

#ifndef _XM_H_
#define _XM_H_

#include <X11/Xlib.h>

/* some type definitions */
typedef struct {
  /* data found in the file */
  unsigned char note;
  char instrument;
  char volume;
  char effect_type;
  char effect_parameter;
} note;

typedef struct {
  /* data found in the file */
/*  0 */  unsigned int sample_length;		/* those three values are bytes length */
/*  4 */  unsigned int sample_loop_start;
/*  8 */  unsigned int sample_loop_length; /* i don't like it. sample_loop_end would have been much better */
/* 12 */  char *sample_data;
/* 16 */  char sample_type;
  char sample_volume;
  char sample_fine_tune;
  unsigned char sample_panning;
  char sample_relative_note_number;
  char sample_name[23];
  /* data for the interface */
  Window window;
  Pixmap pixmap;
  int is_shown;		/* 1 = yes, 0 = no */
  int mode;		/* MOVING, EDIT */
  int subwindow;
  int subsubwin;	/* 0, 1, 2 : what is viewed in the down area (run prog to understand) */
  /* data for subwin 0 (sample viewing) */
  unsigned long max_inc_0;		/* datum */ /* bad name, but in a previous handling, it meant that */
  unsigned long long inc_0;		/* pipo datum */
  unsigned long start_0;		/* datum */
  unsigned long cur_pos_0;		/* datum */
  int down_0;
  unsigned long cut_pos_left;		/* datum */
  unsigned long cut_pos_right;		/* datum */
  int cut_in_use;			/* 0 : no 1 : yes */
  int cur_cut_pos;			/* 0 : left moving 1 : right moving */
  /* subwin 3 (sample bar) */
  int down_3;
  /* subwin 4 (view : sample setting...) */
  int cur_pos_4;
  int down__0_4;
  int down__1_4;
  /* subwin 6 (load) */
  int cur_pos_6;
  /* subwin 7 (save) */
  int cur_pos_7;
  int rate_7;
  /* subwin 0-0 (loop start, end) */
  int cur_pos_0_0;
  /* subwin 0-1 (fine tune, relative tone) */
  int cur_pos_0_1;
  /* subwin 0-2 (volume / panning) */
  int cur_pos_0_2;
  /* subwin 0-3 (8/16b) */
  int mode_3_0;
  int right_0_3;		/* well 3-0 or 0-3, that's the same */
  int cur_pos_0_3;
  /* subwin 0-4 (loop type) */
  int up_0_4;
  int down_0_4;
  int cur_pos_0_4;
  /* subwin 1-0 (record mix rate, 8/16b */
  int record_rate;		/* the rate of recording (8000Hz by default) */
  int record_16b;		/* 0 means 8b 1 16b */
  int cur_pos_1_0;
  /* subwin 1-1 (fragment size, nb fragments) */
  int record_frag_size;
  int record_nb_frags;
  int right_1_1;
  int cur_pos_1_1;
  /* subwin 1-2 (micro or line in) */
  int record_micro_line;		/* 0 for micro 1 for line */
  /* subwin 1-3 (input volume) */
  int record_input_volume;
  int cur_pos_1_3;
  /* subwin 1-4 (go button) */
  int left_1_4;
  int up_1_4;
} sample __attribute__ ((packed));

typedef struct {
  /* data found in the file */
  char name[23];
  short nb_samples;
  char sample_number[96];
  short volume_envelop[12][2];
  short panning_envelop[12][2];
  char nb_volume_points;
  char nb_panning_points;
  char volume_sustain_point;
  char volume_loop_start;
  char volume_loop_end;
  char panning_sustain_point;
  char panning_loop_start;
  char panning_loop_end;
  char volume_type;
  char panning_type;
  char vibrato_type;
  unsigned char vibrato_sweep;
  char vibrato_depth;
  char vibrato_rate;
  unsigned short volume_fadeout;
  sample sample_headers[16];
  /* data for the interface */
  Window window;
  Pixmap pixmap;
  int is_shown;		/* 1 = yes, 0 = no */
  int mode;		/* MOVING, EDIT */
  int subwindow;
  /* data for subwin 0 (volume envelope) */
  int cur_pos_0;
  /* subwin 1 (nb points volume) */
  int cur_pos_1;
  /* subwin 2 (use volume envelope) */
  int cur_pos_2;
  /* subwin 3 (panning envelope) */
  int cur_pos_3;
  /* subwin 4 (nb points for panning) */
  int cur_pos_4;
  /* subwin 5 (use panning envelope) */
  int cur_pos_5;
  /* subwin 6 (fadeout) */
  int cur_pos_6;
  /* subwin 7 (keyboard) */
  int cur_pos_7;
  int up_subwin_7;
  /* subwin 8 (close) */
} instrument;

#include "channels_def.h"

typedef struct {
	    /* data found in the file */
/*    0 */  char module_name[21];
/*   21 */  char tracker_name[21];
/*   42 */  short version;
/*   44 */  int header_size;
/*   48 */  short song_length;
/*   50 */  short restart;
/*   52 */  short nb_channels;
/*   54 */  short nb_patterns;
/*   56 */  short nb_instruments;
/*   58 */  short flags;
/*   60 */  short default_tempo;
/*   62 */  short default_bpm;
/*   64 */  unsigned char patterns_order[256];
/*  320 */  short patterns_nb_rows[256];
/*  832 */  note *patterns[256];
/* 1856 */  instrument *instruments[128];

	    /* data necessary to handle a module in the mixing-process (softly initialised) */
	    /* this var says if we play the module or/and if we play "live" */
/* 2368 */  int mix;
	    /* the new_line function need the following */
	    /* nb_bytes is the length in byte of a tick, speed is initialised to default_tempo then decrease till it reaches 0
	     * then we change current_line, which is the current line in the current pattern which is current_pattern
	     * Clear ? (bpm is the current bpm) (current_tempo is the current tempo (for reinitialisation of bpm))
	     */
/* 2372 */  int nb_bytes;
/* 2376 */  int speed;
/* 2380 */  int bpm;
/* 2384 */  int current_pattern;
/* 2388 */  int current_line;
/* 2392 */  int current_tempo;
	    /* the handle_tick function needs the following */
	    /* to_handle is the number of bytes to handle (it reaches 0 when a tick is over) */
/* 2396 */  int to_handle;
	    /* the current global volume, of course. Init by default to 64 */
/* 2400 */  int global_volume;

	    /* data for the mixer */
	    /* same as to_handle, but for the "live"-player */
/* 2404 */  int live_to_handle;
	    /* if all the ten are used, we must take one away, so do it in circular mode (we take 0 then 1 then ... */
/* 2408 */  int last_live_channel;
	    /* here come the channels */
/* a channel is 284 bytes long */
/* 2412 */  channel channels[32];
	    /* if we dump, then the mixer won't write to the card what it mixes, so there is this boolean */
	    /* =0 if we don't dump 1 otherwise */ /* the mixer put it to 0 when it finishes */
/*11500 */  int do_we_dump;
	    /* number of bytes dumped (because the last step might not be fully dumped) */
/*11504 */  int dumped;
	    /* the channels for the "live"-player (10 should be enough, we have 10 fingers */
/*11508 */  channel live_channels[10];
	    /* and the "live"-pattern. The second note is here to remember what was the last note played */
/* */  note live_notes[10][2];

  /* data for the effects */
  /* those two vars are used to handled B effect (position jump) */
  int next_line_is_a_jump;
  int jump_to;
  /* those for D effect (jump to next pattern, at specified row) */
  /* B and D may mix. Sanx to Ripper for walker.xm */
  int next_line_is_a_set_row;
  int next_row;
  /* for effect e6 */
  int next_line_is_a_loop;
  /* for ee, pattern delay */
  int do_we_delay;
  int delay_turn;

  /* data for the interface */
  /* the window and the pixmap */
  Window window;
  Pixmap pixmap;
  /* mode (MOVING / EDIT) */
  int mode;
  /* size of the pattern subwindow */
  int nb_lines_shown, nb_columns_shown;
  /* when in editing mode, tells in which subwindow we are (logical subwindow, not real, because we have only one window) */
  int subwindow;

  /* subwindow by subwindow, the data necessary (see ../interface/interface.fig) */
  /* 0 (patterns editor) */
  int start_column_0;
  int cursor_column_0;
  int cur_pos_0;			/* 0 = note, 1 = instrument first pos, 2 = instrument 2nd pos,
					   3 = volume 1st pos, 4 = volume 2nd pos, 5 6 7 = effect pos 1 2 3 */
  int up_subwin_0;			/* when we use up arrow, the subwindow where go */

  /* 1 (nb lines shown, nb columns shown) */
  int cur_pos_1;			/* button 0..3, 0 is uppest, 3 is downest */

  /* 2 (bpm, speed) */
  int cur_pos_2;			/* like above */

  /* 3 (song name) */
  /* nothing special */

  /* 4 (instrument/sample arrows) */
  int cur_pos_4;			/* 0..5 up to down */
  int left_subwin_4, right_subwin_4;	/* where go when left/right arrow is pressed */

  /* 5 (instrument) */
  int start_5;				/* number of the 1st shown inst. */
  int cur_pos_5;			/* 0..7 */

  /* 6 (sample) */
  int start_6;
  int cur_pos_6;			/* 0..3 */

  /* 7 (frequence table) */
  int cur_pos_7;
  int up_subwin_7, down_subwin_7;

  /* 8 (sample ed.) */
  /* nothing */

  /* 9 (save) */
  int cur_pos_9;

  /* 10 (inst. ed.) */
  int down_subwin_10;

  /* 11 (close) */
  /* nothing */

  /* 12 (effect menu) */
  /* nothing */

  /* 13 (song length,  ...) */
  int cur_pos_13;			/* 0..7 */
  int left_subwin_13;

  /* 14 (nb chan. / restart pos.) */
  int cur_pos_14;			/* 0..3 */
  int left_subwin_14;

  /* data used by mixer to inform interface that a new line or a new pattern has arrived, so we should redraw things. */
  int new_line;
  int new_pattern;
  /* same but for a new bpm or new tempo (f effect */
  int new_tempo;

  /* the current octave (for inserting notes) */
  int current_octave;			/* 0..7 */

  /* the data to handle cut buffer */
  /* the cut-buffer itself */
  note cut_buffer[256*32];		/* 256 lines, 32 channels, the maximum possible, so everything feets in it */
  int cut_buffer_nb_lines;
  int cut_buffer_nb_columns;
  /* the current selection */
  int current_selection_pattern;	/* the real pattern, not the pattern order */
  int current_selection_nb_lines;
  int current_selection_start_line;
  int current_selection_nb_columns;
  int current_selection_start_column;

  /* we save this xm, so there is a save name. Here it is */
  char *file_name;
} XM __attribute__ ((packed));

/* state of the mixer */
/* 1, 2, 4, 8, ... */
#define MODULE_PLAY	1
#define LIVE_PLAY	2

/* the functions */
int load_xm(char*, XM *);
int save_xm(XM *, char *);
int dump_wav(XM *, char *);
void free_xm(XM *);

#endif /* _XM_H_ */
