/*
 * The Rubick's cube.
 * Had some hours to kill.
 *
 * Sed - april 1999.
 *
 * This is of course total free sofware, or public domain or what you want.
 * Kill copyrights. Kill copylefts.
 *--------------------
 * The screen stuff, ie X. X rules, because of asynchronous stuff via select.
 * threads sucks, just ask the glibc coders to understand.
 */

#include "screen.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "cube.h"
#include "sound.h"
#include "device.h"

/* static int pal[7][3]={{0, 0, 0},
			{63, 0, 0}, 
			{0, 63, 0}, 
			{20, 40, 63}, 
			{63, 63, 0}, 
			{176/4, 48/4, 96/4}, 
			{63, 63, 63}}; */

static int pal[7][3]={{0, 0, 0},
			{63, 0, 0}, 
			{0, 63, 0}, 
			{20, 40, 63}, 
			{63, 63, 0}, 
			{63, 40, 0}, 
			{63, 63, 63}};

static int _8_16[7];
static int _8_32[7];

extern char help_message[];

void set_colormap(SCREEN *s)
{
  XColor color[256];
  int i;

  if (s->depth!=8)
    return;

  for (i=0; i<7; i++) {
    color[i].pixel=i;
    color[i].flags=DoRed|DoGreen|DoBlue;
    color[i].red=((int)pal[i][0])<<10;
    color[i].green=((int)pal[i][1])<<10;
    color[i].blue=((int)pal[i][2])<<10;
  }
  XStoreColors(s->d, s->cm, color, 7);
}

int to16(int col)
{
  return _8_16[col];
}

int to32(int col)
{
  return _8_32[col];
}

/* return -1 if error, 0 if ok */
int init_screen(SCREEN *s)
{
  extern device d;
  XSetWindowAttributes attributs;
  XGCValues gcv;
  XEvent ev;
  XColor a;
  Pixmap p;
  GC pixgc;
  int i;

  for (i=0; i<7; i++) {
    _8_16[i]=((pal[i][0]/2)<<(5+6))+(pal[i][1]<<5)+(pal[i][2]/2);;
  }

  for (i=0; i<7; i++) {
    int r, g, b;
    r=(double)pal[i][0]/63.*255.;
    g=(double)pal[i][1]/63.*255.;
    b=(double)pal[i][2]/63.*255.;
    _8_32[i]=(r<<16)|(g<<8)|b;
  }

  memset(s, 0, sizeof(SCREEN));

  if (!(s->d=XOpenDisplay((char *)0)))
    return -1;

  s->depth=DefaultDepth(s->d, DefaultScreen(s->d));

  if (s->depth!=8 && s->depth!=16 && s->depth!=24) {
    fprintf(stderr, "screen depth not supported (only 8, 16 and 24bpp (which means 32bpp too) handled\n");
    XCloseDisplay(s->d);
    s->d=(Display *)0;
    return -1;
  }

  /* we don't know if s->depth==24 if it's 24 or 32bpp, let's check it */
  if (s->depth==24) {
    int n;
    XPixmapFormatValues *v;

    v=XListPixmapFormats(s->d, &n);
    if (!v || !n) {
      if (v)
	XFree(v);
      fprintf(stderr, "Error while trying to distinguish between 24 or 32bpp, damned...\n");
      XCloseDisplay(s->d);
      s->d=(Display *)0;
      return -1;
    }
    for (n--; n>=0; n--)
      if (v[n].bits_per_pixel==32)
	s->depth=32;
    XFree(v);
  }

  s->buffer=(char *)malloc(320*240/8*s->depth);
  if (!s->buffer) {
    perror("");
    fprintf(stderr, "Meaning no rubix\n");
    XCloseDisplay(s->d);
    s->d=(Display *)0;
    return -1;
  }

  d.depth=s->depth;
  d.buffer=s->buffer;

  if (s->depth==8) {
    s->cm=XCreateColormap(s->d, DefaultRootWindow(s->d), DefaultVisual(s->d, DefaultScreen(s->d)), AllocAll);
    attributs.colormap=s->cm;
  }
  /* let's crate a void cursor for our window */
  p=XCreatePixmap(s->d, DefaultRootWindow(s->d), 1, 1, 1);
  pixgc=XCreateGC(s->d, p, 0, 0);
  a.red=a.green=a.blue=0;
  XFillRectangle(s->d, p, pixgc, 0, 0, 2, 2);
  XFreeGC(s->d, pixgc);
  attributs.cursor=XCreatePixmapCursor(s->d, p, p, &a, &a, 0, 0);
  XFreePixmap(s->d, p);
  /* let's create and map our window */
  if (s->depth==8) {
    s->w=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 320, 240, 3, 8, CopyFromParent, CopyFromParent,
		       CWColormap|CWCursor, &attributs);
  } else {
    s->w=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 320, 240, 3, s->depth==32?24:s->depth,
		       CopyFromParent, CopyFromParent, CWCursor, &attributs);
  }
  {
    XClassHint          xch;

    xch.res_name = "rubix";
    xch.res_class = "Rubix";
    XSetClassHint(s->d, s->w, &xch);
    XStoreName(s->d, s->w, xch.res_name);
  }

  XSelectInput(s->d, s->w, KeyPressMask|ExposureMask|PointerMotionMask/*|FocusChangeMask*/|ButtonPressMask);
  XMapWindow(s->d, s->w);
  do
    XNextEvent(s->d, &ev);
  while (ev.type!=Expose || ev.xexpose.window!=s->w);
  gcv.function=GXcopy;
  s->gc=XCreateGC(s->d,s->w,GCFunction,&gcv);
  /* create our XImage */
  /* bitmap_pad, what the heck is it ?? mail sed@free.fr if you know, please...
   * Well, seems it has to be 32...
   */
  s->im=XCreateImage(s->d, DefaultVisual(s->d, DefaultScreen(s->d)), s->depth==32?24:s->depth, ZPixmap,
		     0, s->buffer, 320, 240, 32, 0);
  if (s->depth==8)
    set_colormap(s);
  /* and finish by moving the pointer to the center of the window */
  place_mouse(s);
  if (XGrabPointer(s->d, s->w, False, PointerMotionMask|ButtonPressMask, GrabModeAsync, GrabModeAsync, s->w, None, None))
    fprintf(stderr, "Could not grab the pointer, the program won't run very nice.\n");
  return 0;
}

/* return -1 if end, 0 otherwise */
int screen_key(SCREEN *s, CUBE *c, SOUND *sd, XKeyEvent *ev)
{
  char dum[50];
  KeySym ks;

  XLookupString(ev, dum, 50, &ks, (XComposeStatus *)0);

  switch(ks) {
    case XK_Escape :
    case XK_q :
    case XK_Q :
      return -1;
    case XK_h :
    case XK_H :
      s->pause=1;
      XUngrabPointer(s->d, 0);
      XFlush(s->d);
      pause_sound(sd);
      {
         char *cmd;

         cmd = malloc(strlen(help_message)+50);
         sprintf(cmd, "xmessage -title \"Rubix: help\" "
                      "-geometry 600x400 \"%s\"", help_message);
         system(cmd);
      }
      usleep(100000);
      XFlush(s->d);
    case XK_space :
      s->pause=1-s->pause;
      if (s->pause) {
	XUngrabPointer(s->d, 0);
	pause_sound(sd);
      } else {
	if (XGrabPointer(s->d, s->w, False, PointerMotionMask|ButtonPressMask, GrabModeAsync, GrabModeAsync, s->w, None, 0))
	  fprintf(stderr, "Could not grab the pointer, the program won't run very nice.\n");
	unpause_sound(sd);
	place_mouse(s);
      }
      break;
    case XK_i :
    case XK_I :
      init_cube(c, NULL, -1);
      display_cube(c, s, sd);
      break;
    case XK_p :
    case XK_P :
      init_cube(c, NULL, 1);
      display_cube(c, s, sd);
      break;
    case XK_r :
    case XK_R :
      reset_cube(c, s);
      display_cube(c, s, sd);
      break;
    case XK_s :
    case XK_S :
      save_cube(c);
      break;
    case XK_m :
    case XK_M :
      if (!s->pause)
	rubick_toggle_sound(sd);
      break;
    case XK_Left :
      if (!s->pause)
	cube_generate_rotate(c, s, sd, 0);
      break;
    case XK_Right :
      if (!s->pause)
	cube_generate_rotate(c, s, sd, 1);
      break;
  }

  return 0;
}

void place_mouse(SCREEN *s)
{
  if (!s->pause) {
    XWarpPointer(s->d, None, s->w, 0, 0, 0, 0, 160, 120);
    XFlush(s->d);
  }
}

void mouse_pre_event(XMotionEvent *ev, int *a, int *b)
{
  if (ev->x==160 && ev->y==120)
    return;

  *a=ev->y-120;
  *b=ev->x-160;
}

void mouse_event(SCREEN *s, CUBE *c, SOUND *sd, int a, int b)
{
#define SPEED (1024.)
  if (s->pause)
    return;

  if (a==0 && b==0)
    return;

  rubick_start_move(sd);
  place_mouse(s);

  /* the x is y and vice-versa */
  cube_rotate(s, c, a*M_PI/SPEED, b*M_PI/SPEED);
  display_cube(c, s, sd);
}

/* return -1 if end, 0 if ok */
int screen_event(SCREEN *s, CUBE *c, SOUND *sd)
{
#define EXPOSE 1
#define MOTION 2
  int type=0;
  int a=0, b=0;
  XEvent ev;
  while(XCheckMaskEvent(s->d, -1L/*KeyPressMask|ExposureMask|PointerMotionMask|FocusChangeMask*/, &ev)!=False) {
    switch(ev.type) {
      case Expose :
	type|=EXPOSE;
	break;
      case KeyPress :
	if (screen_key(s, c, sd, &ev.xkey)==-1)
	  return -1;
	break;
      case ButtonPress :
	if (ev.xbutton.button==Button1) {
	  cube_generate_rotate(c, s, sd, 0);
	} else if (ev.xbutton.button==Button3) {
	  cube_generate_rotate(c, s, sd, 1);
	}
	break;
      case MotionNotify :
	mouse_pre_event(&ev.xmotion, &a, &b);
	type |= MOTION;
	break;
#if 0
      case FocusOut :
	place_mouse(s);
	break;
#endif
    }
  }
  if (type & MOTION) {
    mouse_event(s, c, sd, a, b);
  } else if (type & EXPOSE) {
    put_screen(s);
  }
  return 0;
}

void close_screen(SCREEN *s)
{
  XDestroyWindow(s->d, s->w);
  XFreeGC(s->d, s->gc);
  if (s->depth==8)
    XFreeColormap(s->d, s->cm);
  XCloseDisplay(s->d);
}

void clear_screen(SCREEN *s)
{
  memset(s->buffer, 0, 320*240/8*s->depth);
}

void put_screen(SCREEN *s)
{
  XPutImage(s->d, s->w, s->gc, s->im, 0, 0, 0, 0, 320, 240);
  //XFlush(s->d);
}

/* return ms */
struct timeval get_time(void)
{
  struct timeval t;
  struct timezone d;
  if (gettimeofday(&t, &d)==-1) {
    perror("gettimeofday");
    return t;
  }
  return t;
}
