/*
 * The Rubick's cube.
 * Had some hours to kill.
 *
 * Sed - april 1999.
 *
 * This is of course total free sofware, or public domain or what you want.
 * Kill copyrights. Kill copylefts.
 *--------------------
 * The sound stuff. The player is mine, even if it is not the fastest one, it's mine !
 */

#include "sound.h"

#include "mixer.h"
#include "xm.h"
#include "card.h"
#include "live_player.h"

#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>

/* -1 if bug, 0 if good */
int rubick_init_sound(SOUND *sound, char *file)
{
  int i;

  init_mixer(44100, 16, 1);
  memset(sound, 0, sizeof(SOUND));

  fprintf(stderr, "Initing the sound...\n");
  if (!init_sound(44100, 16, 1)) {
    fprintf(stderr, "Error with initing the sound !\n");
    return 0;
  }
  fprintf(stderr, "Sound inited ok.\n");

  fprintf(stderr, "Loading the module \"%s\"...\n", file);
  if (!load_xm(file, &sound->xm)) {
    fprintf(stderr, "Error with the loading of \"%s\"\n", file);
    close_sound();
    return 0;
  }
  fprintf(stderr, "File \"%s\" loaded ok.\n", file);

  init_xm_data(&sound->xm);

  sound->on=1;
  sound->fd=get_sound_fd();
  sound->xm.mix=MODULE_PLAY;
  sound->xm.mix|=LIVE_PLAY;

  /* not to forget to put void notes in the live player !!! */
  for (i=0; i<10; i++)
    put_live_player_note(&sound->xm, i, 0xff);

  return 0;
}

void rubick_toggle_sound(SOUND *sound)
{
  if (sound->on) {
    close_sound();
    sound->on=0;
    fprintf(stderr, "Sound closed ok\n");
  } else {
    if (sound->xm.mix!=(MODULE_PLAY|LIVE_PLAY)) {
      fprintf(stderr, "Due to init sound error, i am running in no sound mode\n");
      return;
    }
    if (!init_sound(44100, 16, 1)) {
      fprintf(stderr, "Well, problem with re-opening the sound\n");
      return;
    }
    fprintf(stderr, "Sound re-opened ok\n");
    sound->on=1;
    sound->fd=get_sound_fd();
  }
}

void rubick_mix(SOUND *sound)
{
  mix(&sound->xm);
}

void rubick_close_sound(SOUND *sound)
{
  close_sound();
  sound->on=0;
}

unsigned int getmillitime(void)
{
  struct timeval t;
  struct timezone d;
  if (gettimeofday(&t, &d)==-1) {
    perror("gettimeofday");
    return 0;
  }
  return t.tv_sec*1000+t.tv_usec/1000;
}

/* return 1 if in move mode and time over */
int rubick_stop_test(SOUND *s)
{
  unsigned int pt;

  if (!s->on)
    return 0;

  if (!s->move)
    return 0;

  pt=getmillitime();
  if (pt>s->move_time+200)
    return 1;

  /* maybe we've got an overflow ? */
  if (pt < s->move_time)
    return 1;

  return 0;
}

void rubick_start_move(SOUND *s)
{
  int c;

  if (!s->on)
    return;

  if (s->move) {
    s->move_time=getmillitime();
    return;
  }

  s->move=1;
  s->move_time=getmillitime();

  c=get_free_live_player_channel(&s->xm);
  s->move_channel=c;
  s->xm.start_5=MOVE_INSTRUMENT;
  s->xm.cur_pos_5=0;
  put_live_player_note(&s->xm, c, MOVE_NOTE);
}

void rubick_stop_move(SOUND *s)
{
  int c;

  if (!s->on)
    return;

  if (!s->move)
    return;

  s->move=0;

  c=s->move_channel;
  s->xm.start_5=MOVE_INSTRUMENT;
  s->xm.cur_pos_5=0;
  put_live_player_note(&s->xm, c, 96);
}

void rubick_start_rotate(SOUND *s)
{
  int c;

  if (!s->on)
    return;

  c=get_free_live_player_channel(&s->xm);
  s->rotate_channel=c;
  s->xm.start_5=ROTATE_INSTRUMENT;
  s->xm.cur_pos_5=0;
  put_live_player_note(&s->xm, c, ROTATE_NOTE);
}

void rubick_stop_rotate(SOUND *s)
{
  int c;

  if (!s->on)
    return;

  c=s->rotate_channel;
  s->xm.start_5=ROTATE_INSTRUMENT;
  s->xm.cur_pos_5=0;
  put_live_player_note(&s->xm, c, 96);
}

void pause_sound(SOUND *s)
{
  if (!s->on)
    return;

  rubick_toggle_sound(s);
  s->pause=1;
}

void unpause_sound(SOUND *s)
{
  if (!s->pause)
    return;

  s->pause=0;
  rubick_toggle_sound(s);
}
